#include <stdio.h>
#include <assert.h>
#include "opengl.h"
#include "app.h"
#include "gmath/gmath.h"
#include "mesh.h"
#include "meshgen.h"
#include "backdrop.h"
#include "goatvr.h"
#include "opt.h"
#include "fs.h"

static void draw_scene();

int win_width, win_height;
float win_aspect;
long time_msec;
double time_sec;
Mat4 view_matrix;

static bool should_swap;

static float cam_theta, cam_phi;
static float cam_height = 1.65;

static bool bnstate[16];
static int prev_x, prev_y;

static float fov = 60.0;

bool app_init(int argc, char **argv)
{
	if(!init_options(argc, argv, "vrfileman.conf")) {
		return false;
	}
	app_resize(opt.width, opt.height);
	app_fullscreen(opt.fullscreen);

	if(init_opengl() == -1) {
		return false;
	}

	glEnable(GL_MULTISAMPLE);

	int aasamples = 0;
	glGetIntegerv(GL_SAMPLES, &aasamples);
	printf("got %d samples per pixel\n", aasamples);

	printf("Max anisotropy: %d\n", glcaps.max_aniso);

	glEnable(GL_CULL_FACE);
	glEnable(GL_DEPTH_TEST);

	if(opt.srgb && GLEW_ARB_framebuffer_sRGB) {
		printf("enabling sRGB framebuffer\n");
		glEnable(GL_FRAMEBUFFER_SRGB);
	}

	if(opt.vr) {
		if(goatvr_init() == -1) {
			return false;
		}
		goatvr_set_origin_mode(GOATVR_HEAD);

		goatvr_startvr();
		should_swap = goatvr_should_swap() != 0;
		cam_height = goatvr_get_eye_height();
	}

	Mesh::use_custom_sdr_attr = false;

	if(!init_backdrop()) {
		return false;
	}

	if(!init_fs(opt.path)) {
		return false;
	}

	return true;
}

void app_cleanup()
{
	if(opt.vr) {
		goatvr_shutdown();
	}
	cleanup_backdrop();
}

void app_draw()
{
	if(opt.vr) {
		// VR mode
		goatvr_draw_start();
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		for(int i=0; i<2; i++) {
			goatvr_draw_eye(i);

			glMatrixMode(GL_PROJECTION);
			glLoadMatrixf(goatvr_projection_matrix(i, 0.1, 200.0));

			view_matrix = goatvr_view_matrix(i);
			view_matrix.pre_rotate_x(deg_to_rad(cam_phi));
			view_matrix.pre_rotate_y(deg_to_rad(cam_theta));
			view_matrix.pre_translate(0, -cam_height, 0);

			glMatrixMode(GL_MODELVIEW);
			glLoadMatrixf(view_matrix[0]);

			draw_scene();
		}
		goatvr_draw_done();

		if(should_swap) {
			app_swap_buffers();
		}
		app_redraw();	// in VR mode, force continuous redraw

	} else {
		// regular monoscopic mode
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		Mat4 mat;
		mat.perspective(deg_to_rad(fov), win_aspect, 0.1, 200.0);
		glMatrixMode(GL_PROJECTION);
		glLoadMatrixf(mat[0]);

		view_matrix = Mat4::identity;
		view_matrix.pre_rotate_x(deg_to_rad(cam_phi));
		view_matrix.pre_rotate_y(deg_to_rad(cam_theta));
		view_matrix.pre_translate(0, -cam_height, 0);

		glMatrixMode(GL_MODELVIEW);
		glLoadMatrixf(view_matrix[0]);

		draw_scene();

		app_swap_buffers();
		app_redraw();	// since we added animation we need to redisplay even in non-VR mode
	}
	assert(glGetError() == GL_NO_ERROR);
}

static void draw_scene()
{
	draw_backdrop();
	draw_fs();
}

void app_reshape(int x, int y)
{
	glViewport(0, 0, x, y);

	if(opt.vr) {
		goatvr_set_fb_size(x, y, 1.0);
	}
}

void app_keyboard(int key, bool pressed)
{
	if(pressed) {
		switch(key) {
		case 27:
			app_quit();
			break;

		case 'f':
			if(!opt.vr || should_swap) {
				/* we take the need to swap as a signal that our window is not managed
				 * by some VR compositor, and therefore it's safe to fullscreen without
				 * upsetting the VR rendering output
				 */
				opt.fullscreen = !opt.fullscreen;
				app_fullscreen(opt.fullscreen);
			}
			break;

		case ' ':
			if(opt.vr) {
				goatvr_recenter();
			}
			break;

		case '-':
			fov += 1.0;
			if(fov > 160.0) fov = 160.0;
			break;

		case '=':
			fov -= 1.0;
			if(fov < 0.0) fov = 0.0;
			break;
		}
	}
}

void app_mouse_button(int bn, bool pressed, int x, int y)
{
	bnstate[bn] = pressed;
	prev_x = x;
	prev_y = y;
}

void app_mouse_motion(int x, int y)
{
	int dx = x - prev_x;
	int dy = y - prev_y;
	prev_x = x;
	prev_y = y;

	if(!dx && !dy) return;

	if(bnstate[0]) {
		cam_theta += dx * 0.5;

		if(!opt.vr || !goatvr_have_headtracking()) {
			cam_phi += dy * 0.5;

			if(cam_phi < -90) cam_phi = -90;
			if(cam_phi > 90) cam_phi = 90;
		}
	}
	app_redraw();
}
